function last_element(l) { return l[ l.length -1 ] };

// Traditional getElementById wrapper
function $(name) {
    return document.getElementById(name)
}

// Also create $X() for XPath queries
// and $S() for CSS selector queries

// The U combinator. See a real functional programming
// tutorial for the discussion. It allows to create
// non-leaking recursive anonymous functions
function U(f) {
    function __q() {
        args = __q.caller.arguments;
        args.unshift(f);
        return f.call(args);
    };
    return __q
}

// Should add the Y-combinator too, just for coolnes
// and mind-warping

Array.prototype.foreach = function (callback) {
    for (var i=0; i<this.length; i++) {
        callback(this[i]);
    }
}

var debugWindow;
function _debug(msg) {
  return;
  if (! debugWindow) {
    debugWindow = window.open('','debug',"");
    debugWindow.document.write('<html><head><title>Debug</title></head><body><h2>Debug</h2><tt id="debugWindow"></tt></body></html>');
    debugWindow.document.close();
  };
  var w = debugWindow.document.getElementById('debugWindow');
  if (!w)
    alert("No debug window found");
  //alert(w.innerHTML);
  var lines = w.innerHTML.split(/<br>/i);
  while(lines.length > 50)
    lines.shift();
  lines.push(msg);

  w.innerHTML = lines.join("<br />");
};

Function.prototype.method = function (name, func) {
    this.prototype[name] = func;
    return this;
};

// The User class
/*
 * Contains most methods pertaining to editing:
 * - drawing the cursor
 * - moving the cursor
 * - inserting / deleting text
 */

function User (color,name,ofs) {
  this.name = name;
  this.ofs = ofs;
  this.color = color;
  return this;
};

// Rename to setCursor, to keep with JS style
User.method( 'setPosition', function (ofs) {
    this.hide_cursor();
    if (typeof ofs == 'number') {
        this.ofs = ofs;
    };
    
    var i = lines.length -1;
    while (line_offsets[i] > this.ofs) {
        i--;
    };
    
    this.x = this.ofs - line_offsets[i];
    this.y = i;
    this.draw_cursor();
});

// Rename to setCursor, to keep with JS style
User.method( 'set_cursor_xy', function (x,y) {
    if (typeof x == 'number') {
        this.x = x;
        this.y = y;
    };
    
    var ofs = line_offsets[this.y] + this.x;
    this.setPosition(ofs);
});


User.method('cursor_span', function () {
  var c = document.getElementById("cursor_"+this.color);
  if (! c) {
    c = document.createElement("span");
    c.className = "cursor cursor_"+this.color;
    c["id"] = "cursor_"+this.color;
    c.appendChild(document.createTextNode(" "));
  };
  return c
});

// Maybe later prevent/reduce fragmentation by joining
// the text onto
// c.nextSibling or c.prevSibling
// if they have the proper style/attributes.
User.method( 'hide_cursor', function () {
  var c = this.cursor_span();
  var t = document.createTextNode(c.childNodes[0].nodeValue);
  p = c.parentNode;
  if (p) {
    p.replaceChild(t,c);
  };
});

function Change (ofs,old,new_,caret) {
    if ('number' != typeof ofs) {
        new_ = ofs[2];
        old  = ofs[1];
        ofs  = ofs[0];
    };
    this.ofs = ofs;
    this.old = old;
    this.new_ = new_;
    return this;
}

Change.method( 'delta', function () {
    return this.new_.length - this.old.length
});

// should be called toString to be more javascripty
Change.method( 'as_string', function () {
    return "["
         + this.ofs + ","
         + this.encode_string(this.old) + ','
         + this.encode_string(this.new_)
    +"]";
});

Change.method( 'apply', function (str) {
    if (this.old == null) {
        alert("No old in " + this.ofs + " -> " + this.new_);
    }
    return str.substring(0,this.ofs) 
         + this.new_
         + str.substring(this.ofs+this.old.length);
});

Change.method( 'revert', function (str) {
    alert("XXX revert() not implemented");
});

// XXX string.charCodeAt() instead
var char_map = {
    "\\" : "\\\\",
    "\"" : "\\\"",
    "\x00" : "\\x00",
    "\x01" : "\\x01",
    "\x02" : "\\x02",
    "\x03" : "\\x03",
    "\x04" : "\\x04",
    "\x05" : "\\x05",
    "\x06" : "\\x06",
    "\x07" : "\\x07",
    "\x08" : "\\x08",
    "\x09" : "\\x09",
    "\x0a" : "\\x0a",
    "\x0b" : "\\x0b",
    "\x0c" : "\\x0c",
    "\x0d" : "\\x0d",
    "\x0e" : "\\x0e",
    "\x0f" : "\\x0f",
    "\x10" : "\\x10",
    "\x11" : "\\x11",
    "\x12" : "\\x12",
    "\x13" : "\\x13",
    "\x14" : "\\x14",
    "\x15" : "\\x15",
    "\x16" : "\\x16",
    "\x17" : "\\x17",
    "\x18" : "\\x18",
    "\x19" : "\\x19",
    "\x1a" : "\\x1a",
    "\x1b" : "\\x1b",
    "\x1c" : "\\x1c",
    "\x1d" : "\\x1d",
    "\x1e" : "\\x1e",
    "\x1f" : "\\x1f",
}

Change.method( 'encode_string', function (str) {
    return '"' + str.replace(/([\x00-\x1f\\"])/gm, function(m) {
        return char_map[m[0]];
    }) + '"';
});

// Returns a function that will return a merged object
// if the "next" change and this can merge
// Returns undefined if the objects cannot merge
Change.method( 'canMerge', function (next) {
    // alert(this.as_string() + " <- " + next.as_string());
    if (next.ofs == this.ofs + this.new_.length) {
        // next is immediately adjactent to the right
        var c = this;
        return function () {
            return new Change(
                c.ofs,
                c.old + next.old,
                c.new_ + next.new_
            );
        }
    } else if (next.ofs + next.old.length == this.ofs ) {
        // next is immediately adjactent to the left and we shorten
        // the string.
        // XXX This is basically the same as above, except
        // that we need to sort the changes from left to right?
        var c = this;
        return function () {
            return new Change(
                next.ofs,
                next.old + c.old,
                next.new_ + c.new_
            );
        }
    } else {
        return
    }
});

// Class method
// Returns an optimized list of merged changes
// This screams for a version of Array.reduce(),
// that can really return a list instead of a single item
Change.merge = function ( list ) {
    var r = [];
    if (list.length) {
        var i = 1;
        var m = list[0];
        while (i < list.length) {
            var merged;
            if (merged = m.canMerge(list[i])) {
                m = merged();
            } else {
                r.push(m);
                m = list[i];
            };
            i++
        }
        r.push(m);
    };
    return r;
};

// This function only works for merged
// changes! To safeguard against unmerged
// changes, it merges them.
function change_map ( changes ) {
    var result = {};
    changes = Change.merge( changes );
    for (var i = 0; i < changes.length; i++) {
        var c = changes[i];
        var diff = c.delta();
        var ofs = c.ofs + c.old.length;
        if (diff) {
            if (! result[ofs]) {
                result[ofs] = 0;
            };
            var el = [];
            for (var a in result) {
                if ((typeof a) == "string") {
                    el.push(parseInt(a));
                };
            };
            el.sort(function(a,b){return a-b});
            var shift = diff;
            el.foreach(function(affected) {
                shift += result[affected];
                if (affected >= ofs) {
                    result[ affected ] = shift;
                };
            });
        };
    };
    _debug(dump(result));
    return result
}

function adjust_delta(map,ofs) {
    var offsets = [];
    for (var k in map) {
        offsets.push(k);
    }
    offsets = offsets.sort();
    var i = offsets.length-1;
    while (i >= 0 && offsets[i] > ofs) {
        i--
    };
    
    if (i >= 0) {
        return map[offsets[i]]
    } else { 
        return 0
    }
}

function dump(o) {
  var rv = o + "<br>";
  for (p in o) {
    rv = rv + p + ":" + o[p] + " | ";
  };
  return rv
};

// Returns a pair:
// node:   the node containing the C<offset>th char, or null if the 
//         node was not found.
// offset: Remaining offset within the text node
function findNode(top,offset) {
  var r = new Object;
  //_debug("Looking for "+offset);
  r.node = null;
  r.offset = null;

  if (top.nodeValue) {
    if (top.nodeValue.length > offset) {
      //_debug("Found cursor within >"+top.nodeValue+"< ("+offset+"/"+top.nodeValue.length+")");
      r.node = top;
    } else {
      //_debug("Cursor must be after "+top.nodeValue+" ("+offset+"/"+top.nodeValue.length+"); setting to "+(offset - top.nodeValue.length));
      offset = offset - top.nodeValue.length;
    };
  };
  //_debug("Looking below "+top.tagName+" ("+offset+"/?)");
  r.offset = offset;
  for (var i = 0; i < top.childNodes.length; i++) {
    r = findNode(top.childNodes[i], offset);
    //_debug(i+": Got "+r.node);
    if (r.node)
      return r;
    //_debug("Moving to "+r.offset);
    offset = r.offset
  };
  return r;
};

// This is the most ugly code I've written here
// Lots of broken logic. Needs refactoring once
// it works.
// Needs refactoring now, as it doesn't work in IE
// (crashes IE)
User.method( 'draw_cursor', function () {
  $("pos").innerHTML = "Cursor at " + this.x+":"+this.y + "("+this.ofs+")";
  var l = document.getElementById("line_"+this.y);
  if (! l) {
    // nowhere to draw
    return;
  };

  if (! l.childNodes.length) {
    return
  };

  //_debug("Positioning "+this.name);
  var n = findNode(l,this.x);
  if (! n.node) {
    n.node = l.childNodes[ l.childNodes.length -2 ];
    n.offset = n.node.nodeValue.length -1;
  };

  var container = n.node;
  var s = container.nodeValue;
  var ofs = n.offset;

  var firstPortion = s.substring(0,ofs);
  var cursorPortion = s.substring(ofs,ofs+1);
  var lastPortion = s.substring(ofs+1,s.length);

  //_debug(this.x + "/" + this.y + ' split as ' + firstPortion+"|"+cursorPortion+"|"+lastPortion);

  // Update our cursor span
  var c = this.cursor_span();
  c.childNodes[0].nodeValue = cursorPortion;
  var t = document.createTextNode(lastPortion);
  
  // There always is a next sibling, see draw_lines() !
  var nextChild = container.nextSibling;
  
  // XXX Clean up the empty node instead of leaving it in place
  container.nodeValue = firstPortion;  
  l.insertBefore( c, nextChild );
  l.insertBefore( t, nextChild );
});

User.method('at_bol', function () { line_offsets[ this.y ]+1 == this.x });
User.method('at_eol', function () {
  return line_offsets[ this.y+1 ] == this.x;
});

User.method('cursor_down', function () {
    if (this.y >= lines.length -1)
        return;
    this.y += 1;
    this.x = lines[this.y].length < this.x
           ? lines[this.y].length
           : this.x;
    this.set_cursor_xy();
});

User.method('cursor_up', function () {
    if (this.y)
        this.y -= 1;
    this.x = lines[this.y].length < this.x
           ? lines[this.y].length
           : this.x;
    this.set_cursor_xy();
});

User.method('cursor_right', function () {
    // XXX Optionally add no-line-wrap option
    this.setPosition(this.ofs+1);
});

User.method('cursor_left', function () {
    if (this.ofs) {
        this.setPosition(this.ofs-1);
    }
});

User.method('cursor_begin_of_line', function () {
  this.setPosition(line_offsets[this.y]);
});

User.method('cursor_end_of_line', function () {
  this.setPosition(line_offsets[this.y+1]-1);
});

User.method('char_insert', function (ch) {
  do_char_insert(this,ch);
});

User.method('line_join', function (y) {
    // XXX change to delete_char(\n)
    var join_pos = line_offsets[ y+1 ]-1;
    do_delete_char(this,join_pos,1);
});

User.method('char_delete_left', function () {
    this.cursor_left();
    do_delete_char(this);
});

User.method('char_delete_right', function () {
    // XXX change to simple delete_char()
    do_delete_char(this);
});

User.method('name_element', function () {
  var l = document.createElement("li");
  l.className = "user";
  l["id"] = "user_"+this.color;
  l.appendChild( document.createTextNode(this.name));
  return l
});

function draw_lines () {
    e = edit_area();

    // shrink the list
    while (e.childNodes.length > lines.length) {
        e.removeChild( e.childNodes[ e.childNodes.length -1 ]);
    };

    var i = 0;
    while (i < lines.length) {
        var id = "line_" + i;
    
        var old = $(id);
        if (old) {
            // var j = 0;
            // old.normalize(); // Doesn't work with IE
            while (old.childNodes.length > 2) {
                // kill everything after the first child, but keep the span
                old.removeChild(old.childNodes[1]);
            };

            // FireFox Speed hack. Only replace the node if it is different
            if (old.childNodes[0].nodeValue != lines[i]) {
                old.childNodes[0].nodeValue = lines[i];
            };
        } else {
            l = document.createElement("div");
            l.className = "line";
            l["id"] = id;
            l.appendChild(document.createTextNode(lines[i]));

            // Ugly fudge so we always have a last (invisible) element on the line
            eol = document.createElement("span");
            eol.className = "end_of_line";
            eol.appendChild(document.createTextNode(""));
            l.appendChild(eol);
            
            e.appendChild(l);
        };
        i++;
    };
};

function draw_users () {
    var i = 0;

    if (local_user) {
        var a = local_user_area();
        while(a.childNodes.length) {
            a.removeChild(a.childNodes[0]);
        };
        a.appendChild(local_user.name_element());
    };

    users = other_users || new Array;
    
    e = other_user_area();

    // shrink the list
    while (e.childNodes.length > users.length) {
        e.removeChild( e.childNodes[ e.childNodes.length -1 ]);
    };

    while (i < users.length) {
        u = users[i].name_element();
        if (i < e.childNodes.length) {
            e.replaceChild(u, e.childNodes[i]);
        } else {
            e.appendChild(u);
        };
        users[i].setPosition();
        i++;
    };

    if (local_user) {
        local_user.setPosition();
    };
};

function draw_bookmarks () {
    var b = $("uiBookmarks");

    // shrink the list
    while (b.childNodes.length > bookmarks.length) {
        b.removeChild(b.childNodes[ b.childNodes.length -1 ]);
    }

    var i = 0;
    while (i < bookmarks.length) {
        var bm = bookmarks[i].htmlElement();
        if (i < b.childNodes.length) {
            b.replaceChild(bm, b.childNodes[i]);
        } else {
            b.appendChild(bm);
        };
        i++;
    };
};

function edit_area() { return $("codeEdit") };
function other_user_area() { return $("otherUsers") };
function local_user_area() { return $("localUserArea") };

function cmd_cursor_begin_of_line() { local_user.cursor_begin_of_line() };
function cmd_cursor_end_of_line() { local_user.cursor_end_of_line() };
function cmd_cursor_up() { local_user.cursor_up() };
function cmd_cursor_down() { local_user.cursor_down() };
function cmd_cursor_right() { local_user.cursor_right() };
function cmd_cursor_left() { local_user.cursor_left() };
function cmd_char_delete_left() { local_user.char_delete_left() };
function cmd_char_delete_right() { local_user.char_delete_right() };
function cmd_char_newline() { local_user.char_insert("\n"); };

// This is not exactly right for (for example) vi mode
function cmd_char_joinline() { local_user.char_delete_right() };

function stopEvent(e) {
  e.cancelBubble = true;
  e.returnValue = false;

  if (e.stopPropagation)
    e.stopPropagation();

  // IE doesn't have this
  if (e.preventDefault)
    e.preventDefault();
};


// Needed for the navigation keys and any
// special key
function keydownHandler(e) {
  if (! e) { e = window.event };
  var key = e.keyCode;
  
  var s = String.fromCharCode(key);

  k = document.getElementById('keyDisplay');
  k.innerHTML = e.type + " " + key;
  k.innerHTML += e.shiftKey ? " shift" : "";
  k.innerHTML += e.ctrlKey ? " ctrl" : "";
  k.innerHTML += e.altKey ? " alt" : "";
  k.innerHTML += command_map[key];

  if (command_map[key]) {
    command_map[key]();
    stopEvent(e);
    return false;
  };
};

function keyHandler(e) {
  if (! e) { e = window.event };
  var key = e.keyCode;
  var ch = e.charCode != null ? e.charCode : key;
  
  if (e.charCode == null) {
    var m = "";
    for (var p in e) { m += p+": "+e[p] + "\n" };
    // alert(m);
  };
  
  if (e.ctrlKey || e.altKey) { ch = 0; };
  var s = String.fromCharCode(ch);
  
  k = document.getElementById('keyDisplay');
  k.innerHTML = e.type + " " + key;
  k.innerHTML += e.shiftKey ? " shift" : "";
  k.innerHTML += e.ctrlKey ? " ctrl" : "";
  k.innerHTML += e.altKey ? " alt" : "";

  if (ch) {
      // insert key at location
      local_user.char_insert(s);
      stopEvent(e);
      return false;
      
  // If the visual key is bound, execute the command
  } else if (command_map[key]) {
    command_map[key]();
    stopEvent(e);
    return false;
  };
};

function findPos(obj) {
  var pos = new Object;
  pos.left = 0;
  pos.top = 0;
  if (obj.offsetParent)	{
		while (obj.offsetParent) {
		  pos.left += obj.offsetLeft;
			pos.top  += obj.offsetTop;
			obj = obj.offsetParent;
		};
  }	else if (obj.x) {
	  pos.left = obj.x;
	  pos.top = obj.y;
  };
  return pos;
};

function clickPos(e) {
	var pos = new Object;
	pos.left = 0;
	pos.top = 0;
	if (!e) var e = window.event;
	if (e.pageX || e.pageY) {
		pos.left = e.pageX;
		pos.top = e.pageY;
	} else if (e.clientX || e.clientY) {
		pos.left = e.clientX + document.body.scrollLeft;
		pos.top = e.clientY + document.body.scrollTop;
	}
	return pos;
}


function clickHandler(e) {
    // XXX
  // Adjust this by the current viewport scroll position
  // at least for Mozilla!
  // Also adjust this for the font size :(
  var pos = findPos(edit_area());
  var click = clickPos(e);
  x = Math.floor((click.left - pos.left) /10);
  y = Math.floor((click.top - pos.top) /18);

  local_user.set_cursor_xy(x,y);
};

var command_map = new Array;
function initKeymap () {
  command_map[13] = cmd_char_newline;

  command_map[8] = cmd_char_delete_left;
  command_map[46] = cmd_char_delete_right;
  // command_map[100] = cmd_char_delete_right;

  command_map[35] = cmd_cursor_end_of_line;
  command_map[36] = cmd_cursor_begin_of_line;
  command_map[37] = cmd_cursor_left;
  command_map[38] = cmd_cursor_up;
  command_map[39] = cmd_cursor_right;
  command_map[40] = cmd_cursor_down;
};

function Zaphod () {};
Zaphod.Bookmark = function (ofs,id,hotkey,text) {
    this.hotkey = hotkey;
    this.text = text;
    this.id = id;
    this.ofs = ofs;
    this.element = undefined;
    return this;
}

Zaphod.Bookmark.method('setPosition', function (ofs) {
    this.ofs = ofs;
    // this.element = null;
});

Zaphod.Bookmark.method('setText', function (t) {
    this.text = t;
    // this.element = null;
});

Zaphod.Bookmark.method('htmlElement', function () {
    if (! this.element) {
        this.element = document.createElement("li");
        this.element.id = this.id;
        t = document.createTextNode(this.text + "(" + this.ofs + ")");
        this.element.appendChild(t);
    }
    return this.element;
});