#!/usr/bin/perl -w
use strict;
use Test::More tests => 33;
use Data::Dumper;

use_ok 'Zaphod::Editor::Action';

my $expected;
while (<DATA>) {
	chomp;
	if (/^# nok/i) {
		$expected = "";
	} elsif (/^# ok/i) {
		$expected = 'eval $_';
	} elsif (/^#\s*=(.*)/i) {
		diag "Custom value $1";
		$expected = $1;
	};
	next if /^#/;
	my $visual = length $_ >= 23 ? substr($_, 0, 20) . "..." : $_;
	my $e = [eval $expected];
	diag $@ if $@;
    my $g = [Zaphod::Editor::Action::decode($_)];
	is_deeply($g, $e, $visual)
	    or diag Dumper $g, $e;
};

__DATA__
# ok
[1,".","ello"]
[7,"elt","orld"]
[7,"Hello","\"World\"\n"]
[1,"Hello","\"World\"\n"],[2,"Hello","\"World\"\n"],[3,"Hello","\"World\"\n"]

# nok
[1
[1]
[1,
[1,]
[1,"
[1,"]
[1,".
[1,".]
[1,"."
[1,"."]
[1,".",
[1,".",]
[1,".","
[1,".","ello
[1,".","ello"
[1,".","ello",
[1,".","ello",]
[n,".","ello"]
[$foo,".","ello",]
# =[1,"Hello","\"World\"\n"]
[1,"Hello","\"World\"\n"],
[1,"Hello","\"World\"\n"],[
[1,"Hello","\"World\"\n"],[2
[1,"Hello","\"World\"\n"],[2,
[1,"Hello","\"World\"\n"],[2,[3,"Hello","\"World\"\n"]
# =[1,"Hello","\"World\"\n"],[2,"Hello","\"World\"\n"]
[1,"Hello","\"World\"\n"],[2,"Hello","\"World\"\n"],[3,
# =[0,"","a"],[1,"","s"],[2,"","d"],[3,"","q"],[4,"","w"]
[0,"","a"],[1,"","s"],[2,"","d"],[3,"","q"],[4,"","w"]
# =[49,"","\x0a"]
[49,"","\x0a"]